//-----------------------------------------------------------------------------
// EXT_TWI_LM75
// Version 1.2
// 10.11.2011 / UB
// "www.Mikrocontroller-4U.de"
// mc-4u(@)t-online.de
//-----------------------------------------------------------------------------
// Fuer ein externes Modul
// mit dem TWI-IC : LM75
// Temperatur Sensor
//-----------------------------------------------------------------------------
// das File muss im Ordner
// "myAVR/WinAVR/AVR/Include/UB_EXT_MODULE/"
// liegen und mit #include <UB_EXT_MODULE/EXT_TWI_LM75.h>
// eingebunden werden
//-----------------------------------------------------------------------------
// Belegt : TWI-BUS
// Benutzt : MK3_2560_TWI.h = Port-D (Bit0 bis Bit1)
//-----------------------------------------------------------------------------
//
// Info zum LM75
// Grundadresse = 0x90 (wenn keine Pull-UPs aktiv sind)
// 3 Adr-Pins   = Bit1 + Bit2 + Bi3
// Dadurch 8 moegliche Konfigurationen : 90, 92, 94, 96, 98, 9A, 9C, 9E
//
// Aufloesung des Wandlers = 9bit=512 Schritte (0,5C pro Digit)
// KALT    = 0x100 = 256 * 0,5 = -128C
// Mitte-1 = 0x1FF = 1   * 0,5 = -0,5C 
// Mitte   = 0x000 = 0   * 0,5 = 0C
// Mitte+1 = 0x001 = 1   * 0,5 = +0,5C
// HEISS   = 0x0FF = 255 * 0,5 = +127,5C
//-----------------------------------------------------------------------------

//----------------------------------------------------------------------------
// Includes
//----------------------------------------------------------------------------
#include "ub_mk3_2560/MK3_2560_TWI.h" //  bindet die TWI-Funktionen ein


//-----------------------------------------------------------------------------
// INFOS zur Benutzung :
// 01. Funktion "EXT_INIT_LM75" starten
// 02. Funktion "EXT_LM75_READ_TEMP" starten
// 02A. Die Variable : EXT_LM75_OK zeigt an ob OK oder FEHLER
// 02B. Die Temperatur-Daten stehen dann in der Struktur : EXT_LM75
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
// Enumerationen fuer die Funktioen
//-----------------------------------------------------------------------------


typedef enum {
  ADR0_LM75 = 0x90,
  ADR1_LM75 = 0x92,
  ADR2_LM75 = 0x94,
  ADR3_LM75 = 0x96,
  ADR4_LM75 = 0x98,
  ADR5_LM75 = 0x9A,
  ADR6_LM75 = 0x9C,
  ADR7_LM75 = 0x9E
}ADR_LM75;


//-----------------------------------------------------------------------------
// Die gelesenen Temperatur-Werte werden in einer Struktur uebergeben
// um im externen Programm keine Umwandlung machen zu muessen
// Bsp +23,5C -> raw_temp=+23.5 / negativ=false / abs_temp=23 / rest=5
// Bsp -12,0C -> raw_temp=-12.0 / negativ=true  / abs_temp=12 / rest=0
//-----------------------------------------------------------------------------
typedef struct {
  float raw_temp;	// Temp-Wert als Kommazahl von -128,0C bis +127,5C
  bool negativ;         // true falls die Temperatur kleiner 0C
  char abs_temp;	// absolutwert der Temperatur (ohne Vorzeichen und als Ganzzahl)
  char rest;		// Nachkommastelle von abs_temp
}Temp_t;

//-----------------------------------------------------------------------------
// Funktionen die vom MAIN benutzt werden koennen
//-----------------------------------------------------------------------------
void EXT_INIT_LM75(void);
void EXT_LM75_READ_TEMP(ADR_LM75 twi_adr);


//-----------------------------------------------------------------------------
// Globale Variabeln
//-----------------------------------------------------------------------------
bool EXT_LM75_OK=false;       	// zum ueberpruefen der fehlerfreien Funktion
Temp_t EXT_LM75;		// hier stehen die Temp-Istwerte drinn

//-----------------------------------------------------------------------------
// Defines
//-----------------------------------------------------------------------------
#define LM75_REG_TEMP 0x00	// Register-Adresse der Temperatur


//-----------------------------------------------------------------------------
// interne Funktionen
//-----------------------------------------------------------------------------
void LM75_MAKE_TEMP(char hi, char lo);



//-----------------------------------------------------------------------------
void EXT_INIT_LM75(void) {	// zum initialisieren des Temp-Sensors

  // init vom TWI-BUS
  MK3_INIT_TWI();

  // init der Struktur ... auf 0,0C
  EXT_LM75.raw_temp=0.0;
  EXT_LM75.negativ=false;
  EXT_LM75.abs_temp=0;
  EXT_LM75.rest=0;
}


//-----------------------------------------------------------------------------
void EXT_LM75_READ_TEMP(ADR_LM75 twi_adr) {
  // auslesn des LM75 und auffuellen der Temp_Struktur
  // twi_adr ist die Adresse vom externen IC 
  // 
  // bei einem Fehler wird EXT_24C02_OK auf FALSE gesetzt sonst auf TRUE

  bool ok=false;
  char antwort=0;
  char lo_byte=0,hi_byte=0;

  EXT_LM75_OK=true; // zuerst auf OK

  ok=MK3_TWI_START();	        // sendet Start-Kennung
  if(ok==true) {
    // nur wenn Startkennung erfolgreich
    antwort=MK3_TWI_SEND(twi_adr,false); // 1.Byte = Slave-Adresse (ohne ACK)
    if(antwort==TWSR_SLAVE_OK) {
      // Slave hat mit ACK geantwortet
      MK3_TWI_SEND(LM75_REG_TEMP,false); // 2.Byte = Register_0 (ohne ACK)
      MK3_TWI_START();		         // sendet Start-Kennung
      MK3_TWI_SEND(twi_adr+1,false);	 // 1.Byte = Slave-Adresse+1 (ohne ACK)
      hi_byte=MK3_TWI_READ(true);        // 2.Byte wird vom Slave gesendet (mit ACK)
      lo_byte=MK3_TWI_READ(false);       // 3.Byte wird vom Slave gesendet (ohne ACK)

      // ausrechnen der Temperatur
      LM75_MAKE_TEMP(hi_byte,lo_byte);
    }
    else {
      // Slave hat nicht geantwortet 
      EXT_LM75_OK=false;
    }
    MK3_TWI_STOP();		         // sendet Stop-Kennung
  }
  else {
    // Startkennung falsch
    EXT_LM75_OK=false;
  }
}


//-----------------------------------------------------------------------------
void LM75_MAKE_TEMP(char hi, char lo) {
  // rechnet aus Lo- und Hi-Byte die Temperatur aus

  char dummy=0;

  if((hi&0x80)==0) {
    // die Zahl ist positiv
    EXT_LM75.negativ=false;

    if((lo&0x80)==0) {       // lo ist der Rest
      // komma = 0
      EXT_LM75.rest=0;
      EXT_LM75.abs_temp=hi;    // hi ist der Temp-wert
      EXT_LM75.raw_temp=hi;
    }
    else {
      // komma = 5
      EXT_LM75.rest=5;
      EXT_LM75.abs_temp=hi;
      EXT_LM75.raw_temp=hi+0.5;
    }
  }
  else {
    // die Zahl ist negativ
    EXT_LM75.negativ=true;

    dummy=0xFF^hi;           // hi ist der Komplementaer Temp-wert
    if((lo&0x80)==0) {
      // komma = 0
      dummy++;               // Zweierkomplement
      EXT_LM75.rest=0;
      EXT_LM75.abs_temp=dummy;
      EXT_LM75.raw_temp=0-dummy;
        
    }
    else {
      // komma = 5
      EXT_LM75.rest=5;
      EXT_LM75.abs_temp=dummy;
      EXT_LM75.raw_temp=0-dummy-0.5;
    }
  }
}



